import scrapy
from scrapy.crawler import AsyncCrawlerProcess


class NoRequestsSpider(scrapy.Spider):
    name = "no_request"

    async def start(self):
        return
        yield


process = AsyncCrawlerProcess(settings={})

process.crawl(NoRequestsSpider)
process.crawl(NoRequestsSpider)
process.start()
